type typ = Note | Void | Int | String | Float | Bool | Rhythm | Pitch | Array of typ

type op = Add | Sub | Mult | Div | Mod | Equal | Neq | Less | Leq | Greater | Geq |
          And | Or

type uop = Neg | Not
type prop = R | P

type bind = typ * string

type expr = 
    PLiteral of string
  | StrLiteral of string
  | RLiteral of string
  | ArrayLiteral of expr list
  | Call of string * expr list
  | Noexpr
  | Literal of int
  | Binop of expr * op * expr
  | Unop of uop * expr
  | Fliteral of string
  | BoolLit of bool
  | Assign of expr * expr
  | Id of string
  | CreateNote of expr list
  | NoteAccess of expr * prop
  | NoteAssign of expr * prop * expr
  | CreateArray of typ * expr 
  | ArrayAccess of expr * expr
  | ArrayLength of expr


type statement =
    Expr of expr
  | Block of statement list
  | Return of expr
  | Bind of bind
  | BindAssign of typ * string * expr
  | If of expr * statement * statement
  | For of statement * expr * expr * statement
  | While of expr * statement

type func_decl = {
    typ : typ;
    fname : string;
    formals : bind list;
    body : statement list;
  }

type program = statement list * func_decl list

(* Pretty-printing functions *)

let string_of_op = function
    Add -> "+"
  | Sub -> "-"
  | Mult -> "*"
  | Div -> "/"
  | Mod -> "%"
  | Equal -> "=="
  | Neq -> "!="
  | Less -> "<"
  | Leq -> "<="
  | Greater -> ">"
  | Geq -> ">="
  | And -> "&&"
  | Or -> "||"

let string_of_uop = function
    Neg -> "-"
  | Not -> "!"

let rec string_of_typ = function
    Int -> "int"
  | Bool -> "bool"
  | Float -> "float"
  | Void -> "void"
  | String -> "string"
  | Note -> "note"
  | Rhythm -> "rhythm"
  | Pitch -> "pitch"
  | Array t -> string_of_typ t ^ "[]"

let string_of_prop = function
    R -> "rhythm"
  | P -> "pitch"

let rec string_of_expr = function
    PLiteral(l) -> l
    |     Literal(l) -> string_of_int l
    |     StrLiteral(l) ->  "\""^l^"\""
  | RLiteral(l) -> l
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | CreateNote(el) ->
      "new Note(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | Noexpr -> ""
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_expr e2
  | Unop(o, e) -> string_of_uop o ^ string_of_expr e
  | Fliteral(l) -> l
  | BoolLit(true) -> "true"
  | BoolLit(false) -> "false"
  | Assign(v, e) -> string_of_expr v ^ " = " ^ string_of_expr e
  | Id(s) -> s
  | NoteAccess(s, p) -> string_of_expr s ^ "." ^ string_of_prop p
  | NoteAssign(s, p, e) -> string_of_expr s ^ "." ^ string_of_prop p ^ " = " ^ string_of_expr e
  | ArrayLiteral(l) -> "[" ^ (string_of_list (List.map string_of_expr l)) ^ "]"
  | ArrayAccess(s, e) -> string_of_expr s ^ "[" ^ string_of_expr e ^ "]"
  | ArrayLength(l) -> "length of " ^ string_of_expr l (*TODO: make this better *)
  | CreateArray(t, e) -> "new " ^ string_of_typ t ^ "[" ^ string_of_expr e ^ "]"

and string_of_list l =
  let add_comma a item =
    a ^ item ^"," in
  List.fold_left add_comma "" l

let rec string_of_stmt = function
    Expr(expr) -> string_of_expr expr ^ ";\n"
  |     Block(stmts) ->
          "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n"
  | Bind(t, id) -> string_of_typ t ^ " " ^ id ^ ";\n"
  | BindAssign(t, id, e) -> string_of_typ t ^ " " ^ id ^ " = " ^ string_of_expr e ^ ";\n"
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | For(s1, e2, e3, s) ->
    "for (" ^ string_of_stmt s1 ^ string_of_expr e2 ^ " ; " ^
    string_of_expr e3  ^ ") " ^ string_of_stmt s
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") " ^ string_of_stmt s

let string_of_vdecl (t, id) = string_of_typ t ^ " " ^ id ^ ";\n"

let string_of_fdecl fdecl =
  string_of_typ fdecl.typ ^ " " ^
  fdecl.fname ^ "(" ^ String.concat ", " (List.map snd fdecl.formals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_stmt fdecl.body) ^
  "}\n"

let string_of_program (stmts, funcs) =
  String.concat "" (List.map string_of_stmt (List.rev stmts)) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
