module L = Llvm
module A = Ast
open Sast

module StringMap = Map.Make(String)

let translate (statements, functions) =
  let context    = L.global_context () in
  
  (* Create the LLVM compilation module into which
     we will generate code *)
let the_module = L.create_module context "E-CATZ" in

  (* Get types from the context *)
let i32_t      = L.i32_type    context
and i8_t   = L.i8_type     context
and void_t = L.void_type   context
and i1_t       = L.i1_type     context
and float_t    = L.double_type context
and string_t   = L.pointer_type (L.i8_type context)
in
let pitch_t    = string_t
and rhythm_t   = string_t
in
let note_t     = L.struct_type context [| pitch_t; rhythm_t |]
    in
let rec ltype_of_typ = function
      A.String   -> string_t
    | A.Note -> note_t
    | A.Void -> void_t
    | A.Int  -> i32_t
    | A.Bool  -> i1_t
    | A.Float -> float_t
    | A.Rhythm -> rhythm_t
    | A.Pitch -> pitch_t
    | A.Array t ->  array_t t
and array_t t = L.struct_type context [| L.pointer_type (ltype_of_typ t); i32_t |]
    in

(* all of our beautiful built-in c functions *)
  let printf_t : L.lltype =
      L.var_arg_function_type i32_t [| L.pointer_type i8_t |] in
  let printf_func : L.llvalue =
      L.declare_function "printf" printf_t the_module in

let write_t: L.lltype =
      L.var_arg_function_type i32_t [| array_t A.Note ; string_t |] in
  let write_func : L.llvalue =
      L.declare_function "write" write_t the_module in

let writeN_t: L.lltype =
      L.var_arg_function_type i32_t [| array_t (A.Array A.Note) ; string_t |] in
  let writeN_func : L.llvalue =
      L.declare_function "writeN" writeN_t the_module in

let randInt_t: L.lltype =
      L.var_arg_function_type i32_t [| i32_t; i32_t |] in
  let randInt_func : L.llvalue =
      L.declare_function "randInt" randInt_t the_module in

let seed_t: L.lltype =
      L.var_arg_function_type i32_t [| |] in
  let seed_func : L.llvalue =
      L.declare_function "seed" seed_t the_module in

(* put all code that isn't inside a function inside a big dummy main function *)
    let build_main stmts funcs=
        let main : L.lltype =
            L.var_arg_function_type i32_t [| |] in
        let main_func : L.llvalue =
            L.define_function "main" main the_module in
        let builder = L.builder_at_end context (L.entry_block main_func) in


  (* Format to String *)
  let int_format_str = L.build_global_stringptr "%d\n" "fmt" builder
  and float_format_str = L.build_global_stringptr "%g\n" "fmt" builder 
  and str_format_str = L.build_global_stringptr "%s\n" "fmt" builder in

 (* Define each function (arguments and return type) so we can
     call it even before we've created its body *)
     let function_decls : (L.llvalue * sfunc_decl) StringMap.t =
      let function_decl m fdecl =
        let name = fdecl.sfname
        and formal_types =
    Array.of_list (List.map (fun (t,_) -> ltype_of_typ t) fdecl.sformals)
        in let ftype = L.function_type (ltype_of_typ fdecl.styp) formal_types in
        StringMap.add name (L.define_function name ftype the_module, fdecl) m in
      List.fold_left function_decl StringMap.empty functions in

      (* because of how data is stored and loaded into arrays
        this function allows you to get data in ids and in array brackets (x[3])
        without directly loading it. When this function is called
        another piece of the code knows it's responsible for loading it later.
        We were having issues with double loads *)
      let rec access builder locals ((t, e) : sexpr) =
      match e with
          SId s       -> StringMap.find s locals
        | SArrayAccess(s, e) ->
            let array_struct = access builder locals s in
            let array_ptr = L.build_struct_gep array_struct 0 "array_ptr" builder in
            let array = L.build_load array_ptr "array" builder in
            let index = expr builder locals e in
            let element_ptr = L.build_gep array [| index |] "array_access" builder in
            element_ptr
        | _ -> raise(Failure("This expression " ^ string_of_sexpr (t, e) ^ " does not access data."))

      (* This function was created for the same reason as above except for when dealing with Note
        structs instead. Again a double load problem. Also, it makes so that pitch literals actually
        act as pitches instead of being a note constructor. This is only used when doing dot access to a
        note. Additionally, if it doesn't get the data it wants,
        it just makes a C4 note instead of throwing an error *)
      and get_needed_struct_value builder locals ((_, e) : sexpr) =
      match e with
          SPLiteral p -> L.build_global_stringptr p "pitch" builder
        | SNoteAccess(s, A.P) ->
            let note_struct = access builder locals s in
            let element_ptr = L.build_struct_gep note_struct 0 "element_ptr" builder in
            L.build_load element_ptr "element" builder
        | SNoteAccess(s, A.R) ->
            let note_struct = access builder locals s in
            let element_ptr = L.build_struct_gep note_struct 1 "element_ptr" builder in
            L.build_load element_ptr "element" builder
        | SRLiteral r -> L.build_global_stringptr r "rhythm" builder
        | SNoteAssign(s, p, e) ->
            let note_struct = access builder locals s in
            let idx = (match p with
                A.P -> 0
              | A.R -> 1
            ) in
            let element_ptr = L.build_struct_gep note_struct idx "element_ptr" builder in
            let value = get_needed_struct_value builder locals e in
            ignore(L.build_store value element_ptr builder); value
        | _ -> L.build_global_stringptr "C4" "pitch" builder


      (* Construct code for an expression; return its value *)
      and expr builder locals ((_, e) : sexpr) =
      match e with
          SPLiteral p ->
            (* a pitch literal on its own acts as a note constructor*)
            let note_ptr = L.build_malloc note_t "note_ptr" builder in
            let pitch_ptr = L.build_struct_gep note_ptr 0 "pitch_ptr" builder in
            let pitch = L.build_global_stringptr p "pitch" builder in
            ignore(L.build_store pitch pitch_ptr builder);
            let rhythm_ptr = L.build_struct_gep note_ptr 1 "rhythm_ptr" builder in
            (* default quarter note rhythm *)
            let rhythm = expr builder locals ((A.Rhythm, SRLiteral("qr"))) in
            ignore(L.build_store rhythm rhythm_ptr builder);
            L.build_load note_ptr "note" builder
        | SLiteral i  -> L.const_int i32_t i
        | SRLiteral r -> L.build_global_stringptr r "rhythm" builder
        | SStrLiteral a -> L.build_global_stringptr a "str" builder
        | SNoexpr     -> L.const_int i32_t 0
        | SBoolLit b  -> L.const_int i1_t (if b then 1 else 0)
        | SFliteral l -> L.const_float_of_string float_t l
        | SId s       -> L.build_load (StringMap.find s locals) s builder
      | SAssign (s, e) -> let e' = expr builder locals e in
                          ignore(L.build_store e' (access builder locals s) builder); e'
      | SCall ("write", [e1; e2]) ->
      L.build_call write_func [| (expr builder locals e1); (expr builder locals e2) |]
        "write" builder
      | SCall ("writeN", [e1; e2]) ->
            L.build_call writeN_func [| (expr builder locals e1); (expr builder locals e2) |]
              "writeN" builder
      | SCall ("print", [e]) | SCall ("printb", [e])  ->
            L.build_call printf_func [| int_format_str; (expr builder locals e ) |]
            "printf" builder
      | SCall ("prints", [e]) ->
            L.build_call printf_func [| str_format_str ; (expr builder locals e) |] 
            "printf" builder
      | SCall ("printf", [e]) -> 
            L.build_call printf_func [| float_format_str ; (expr builder locals e) |]
              "printf" builder
      | SCall ("randInt", [e1; e2])  ->
          L.build_call randInt_func [| (expr builder locals e1 );  (expr builder locals e2 ) |]
                "randInt" builder
      | SCall ("seed", [])  ->
          L.build_call seed_func [| |]
                "seed" builder
      | SCall (f, args) ->
         let (fdef, fdecl) = StringMap.find f function_decls in
         let llargs = List.rev (List.map (expr builder locals) (List.rev args)) in
         let result = (match fdecl.styp with
                              A.Void -> ""
                            | _ -> f ^ "_result") in
               L.build_call fdef (Array.of_list llargs) result builder
      | SCreateNote(args) ->
          let note_ptr = L.build_malloc note_t "note_ptr" builder in
          let pitch_ptr = L.build_struct_gep note_ptr 0 "pitch_ptr" builder in
          let pitch = get_needed_struct_value builder locals (List.hd args) in
          ignore(L.build_store pitch pitch_ptr builder);
          let rhythm_ptr = L.build_struct_gep note_ptr 1 "rhythm_ptr" builder in
          let rhythm = expr builder locals (List.nth args 1) in
          ignore(L.build_store rhythm rhythm_ptr builder);
          L.build_load note_ptr "note" builder
      | SUnop(op, ((t, _) as e)) ->
          let e' = expr builder locals e in
        (match op with
          A.Neg when t = A.Float -> L.build_fneg 
        | A.Neg                  -> L.build_neg
        | A.Not                  -> L.build_not) e' "tmp" builder
      | SNoteAccess(s, p) ->
          let note_struct = access builder locals s in
          let idx = (match p with
              A.P -> 0
            | A.R -> 1
          ) in
          let element_ptr = L.build_struct_gep note_struct idx "element_ptr" builder in
          L.build_load element_ptr "element" builder
      | SNoteAssign(s, p, e) ->
          let note_struct = access builder locals s in
          let idx = (match p with
              A.P -> 0
            | A.R -> 1
          ) in
          let element_ptr = L.build_struct_gep note_struct idx "element_ptr" builder in
          let value = get_needed_struct_value builder locals e in
          ignore(L.build_store value element_ptr builder); value
      | SCreateArray(t, l) ->
          (* create an empty array of desired length *)
          let l' = expr builder locals l in
          let t' = ltype_of_typ t in
          let struct_ptr = L.build_malloc (array_t t) "struct_ptr" builder in
          let array_in_struct_ptr = L.build_struct_gep struct_ptr 0 "array_in_struct_ptr" builder in
          let length_in_struct_ptr = L.build_struct_gep struct_ptr 1 "length_in_struct_ptr" builder in
          ignore(L.build_store l' length_in_struct_ptr builder);
          let array_ptr = L.build_array_malloc t' l' "array_lit" builder
          in
          ignore(L.build_store array_ptr array_in_struct_ptr builder);
          L.build_load struct_ptr "struct" builder
      | SArrayAccess(s, e) ->
          let array_struct = access builder locals s in
          let array_ptr = L.build_struct_gep array_struct 0 "array_ptr" builder in
          let array = L.build_load array_ptr "array" builder in
          let index = expr builder locals e in
          let element_ptr = L.build_gep array [| index |] "array_access" builder in
          L.build_load element_ptr "element" builder
      | SArrayLength(s) ->
         let array_struct = access builder locals s in
         let length_ptr = L.build_struct_gep array_struct 1 "length_ptr" builder in
         L.build_load length_ptr "length" builder
      | SArrayLiteral(e) ->
          let length = L.const_int i32_t (List.length e) in
          let t = fst (List.hd e) in
          let t' = ltype_of_typ t in
          let struct_ptr = L.build_malloc (array_t t) "struct_ptr" builder in
          let array_in_struct_ptr = L.build_struct_gep struct_ptr 0 "array_in_struct_ptr" builder in
          let length_in_struct_ptr = L.build_struct_gep struct_ptr 1 "length_in_struct_ptr" builder in
          ignore(L.build_store length length_in_struct_ptr builder);
          let array_ptr = L.build_array_malloc t' length "array_lit" builder
          in
          (* iterate through each element in the array literal and build it *)
          let build_element idx e1 =
            let e1' = expr builder locals e1 in
            let llindex = L.const_int i32_t idx in
            let element_ptr = L.build_gep array_ptr [| llindex |] "build_elem" builder in
            let casted_element_ptr = L.build_pointercast element_ptr (L.pointer_type t') "casted" builder in
            ignore(L.build_store e1' casted_element_ptr builder);
          in ignore(List.iteri build_element e);
          ignore(L.build_store array_ptr array_in_struct_ptr builder); L.build_load struct_ptr "struct" builder
      | SBinop ((A.Float,_ ) as e1, op, e2) ->
          let e1' = expr builder locals e1
          and e2' = expr builder locals e2  in
          (match op with
            A.Add     -> L.build_fadd
          | A.Sub     -> L.build_fsub
          | A.Mult    -> L.build_fmul
          | A.Div     -> L.build_fdiv
          | A.Equal   -> L.build_fcmp L.Fcmp.Oeq
          | A.Neq     -> L.build_fcmp L.Fcmp.One
          | A.Less    -> L.build_fcmp L.Fcmp.Olt
          | A.Leq     -> L.build_fcmp L.Fcmp.Ole
          | A.Greater -> L.build_fcmp L.Fcmp.Ogt
          | A.Geq     -> L.build_fcmp L.Fcmp.Oge
          | A.And | A.Or | A.Mod ->
              raise (Failure "internal error: semant should have rejected and/or/mod on float")
          ) e1' e2' "tmp" builder
            | SBinop (e1, op, e2) ->
          let e1' = expr builder locals e1
          and e2' = expr builder locals e2  in
          (match op with
            A.Add     -> L.build_add
          | A.Sub     -> L.build_sub
          | A.Mult    -> L.build_mul
          | A.Div     -> L.build_sdiv
          | A.Mod     -> L.build_srem
          | A.And     -> L.build_and
          | A.Or      -> L.build_or
          | A.Equal   -> L.build_icmp L.Icmp.Eq
          | A.Neq     -> L.build_icmp L.Icmp.Ne
          | A.Less    -> L.build_icmp L.Icmp.Slt
          | A.Leq     -> L.build_icmp L.Icmp.Sle
          | A.Greater -> L.build_icmp L.Icmp.Sgt
          | A.Geq     -> L.build_icmp L.Icmp.Sge
          ) e1' e2' "tmp" builder

      in

      (* LLVM insists each basic block end with exactly one "terminator"
         instruction that transfers control.  This function runs "instr builder"
         if the current block does not already have a terminator.  Used,
         e.g., to handle the "fall off the end of the function" case. *)
      let add_terminal builder instr =
        match L.block_terminator (L.insertion_block builder) with
    Some _ -> ()
        | None -> ignore (instr builder) in

    let add_local m (t, n) p b =
          L.set_value_name n p;
    let local = L.build_alloca (ltype_of_typ t) n b in
          ignore (L.build_store p local b);
    StringMap.add n local m

    in

    (* this function is used to make it so we can maintain our master locals map
      of variables *)
    let rec build_stmt_list ftyp func builder locals = function
          hd :: tl -> let (b, l) = build_stmt ftyp func builder locals hd in
            build_stmt_list ftyp func b l tl
        | [] -> (builder, locals)

    and build_stmt ftyp func builder locals = function
          SBlock sl -> build_stmt_list ftyp func builder locals sl
        | SExpr e -> ignore(expr builder locals e ); (builder, locals)
        | SReturn e -> ignore(match ftyp with
                                  A.Void -> L.build_ret_void builder
                                | _ -> L.build_ret (expr builder locals e ) builder );
                                 (builder, locals)
         | SIf (predicate, then_stmt, else_stmt) ->
         let bool_val = expr builder locals predicate in
	 let merge_bb = L.append_block context "merge" func in
         let build_br_merge = L.build_br merge_bb in (* partial function *)

	 let then_bb = L.append_block context "then" func in
	 add_terminal ( fst (build_stmt ftyp func (L.builder_at_end context then_bb) locals then_stmt))
	   build_br_merge;

	 let else_bb = L.append_block context "else" func in
	 add_terminal (fst (build_stmt ftyp func (L.builder_at_end context else_bb) locals else_stmt))
	   build_br_merge;

	 ignore(L.build_cond_br bool_val then_bb else_bb builder);
	 (L.builder_at_end context merge_bb, locals) 
        | SBindAssign (ty, name, sexpr) -> let x = expr builder locals sexpr in
                                           let locals' = add_local locals (ty, name) x builder in
                                           (builder, locals')
        | SBind (ty, name) -> let locals' = add_local locals (ty, name) (L.const_int i32_t 0) builder in (builder, locals')
        | SWhile (predicate, body) ->
          let pred_bb = L.append_block context "while" func in
          ignore(L.build_br pred_bb builder);
      
          let body_bb = L.append_block context "while_body" func in
          add_terminal (fst (build_stmt ftyp func (L.builder_at_end context body_bb) locals body)) (* (stmt (L.builder_at_end context body_bb) body) *)
            (L.build_br pred_bb);
      
          let pred_builder = L.builder_at_end context pred_bb in
          let bool_val = expr pred_builder locals predicate in
      
          let merge_bb = L.append_block context "merge" func in
          ignore(L.build_cond_br bool_val body_bb merge_bb pred_builder);
          (L.builder_at_end context merge_bb, locals)
      
            (* Implement for loops as while loops *)
          | SFor (s1, e2, e3, body) -> let (b, l) = build_stmt ftyp func builder locals s1 in
          build_stmt ftyp func b l
            (  SWhile (e2, SBlock [body ; SExpr e3]) )
    in

        (* this is the code that actually starts to build statements that are not inside of functions
          it starts with an empty locals table that is populated as it reads Binds in check_stmt*)
        let (builder, _) =
            build_stmt A.Int main_func builder StringMap.empty (SBlock stmts);
        in
        add_terminal builder (L.build_ret (L.const_int i32_t 0));

        (* Fill in the body of the given function *)
        let build_function_body fdecl =
          let (the_function, _) = StringMap.find fdecl.sfname function_decls in
          let builder = L.builder_at_end context (L.entry_block the_function) in

       (* Construct the function's "locals": formal arguments and locally
             declared variables.  Allocate each on the stack, initialize their
             value, if appropriate, and remember their values in the "locals" map *)
          let local_vars =
            let add_formal m (t, n) p =
              L.set_value_name n p;
        let local = L.build_alloca (ltype_of_typ t) n builder in
              ignore (L.build_store p local builder);
        StringMap.add n local m
            in

          List.fold_left2 add_formal StringMap.empty fdecl.sformals
              (Array.to_list (L.params the_function)) in

         (* Build the code for each statement in the function *)
          let (builder, _) = build_stmt fdecl.styp the_function builder local_vars (SBlock fdecl.sbody)  in

          (* Add a return if the last block falls off the end *)
          add_terminal builder (match fdecl.styp with
              A.Void -> L.build_ret_void
            | A.Float -> L.build_ret (L.const_float float_t 0.0)
            | t -> L.build_ret (L.const_int (ltype_of_typ t) 0))
        in
        List.iter build_function_body funcs
    in

    build_main statements functions;
    the_module

