(* Semantically-checked Abstract Syntax Tree and functions for printing it *)

open Ast

type sexpr = typ * sx 
and sx = 
	  SPLiteral of string
  | SStrLiteral of string
  | SLiteral of int
  | SRLiteral of string
  | SFliteral of string
  | SBoolLit of bool
  | SArrayLiteral of sexpr list
  | SCall of string * sexpr list
  | SNoexpr
  | SBinop of sexpr * op * sexpr
  | SUnop of uop * sexpr
  | SAssign of sexpr * sexpr
  | SId of string
  | SCreateNote of sexpr list
  | SNoteAccess of sexpr * prop
  | SNoteAssign of sexpr * prop * sexpr
  | SCreateArray of typ * sexpr
  | SArrayAccess of sexpr * sexpr
  | SArrayLength of sexpr

type sstatement =
    SExpr of sexpr
  | SBlock of sstatement list
  | SReturn of sexpr
  | SBind of bind
  | SBindAssign of typ * string * sexpr
  | SIf of sexpr * sstatement * sstatement
  | SFor of sstatement * sexpr * sexpr * sstatement
  | SWhile of sexpr * sstatement

type sfunc_decl = {
    styp : typ;
    sfname : string;
    sformals : bind list;
    sbody : sstatement list;
  }

type sprogram = sstatement list * sfunc_decl list

(* Pretty-printing functions *)

let rec string_of_sexpr (t, e) =
  "(" ^ string_of_typ t ^ " : " ^ (match e with
    SPLiteral(l) -> l
  | SStrLiteral(l) -> "\"" ^ l ^ "\""
  | SLiteral(l) -> string_of_int l
  | SRLiteral(l) -> l
  | SBoolLit(true) -> "true"
  | SBoolLit(false) -> "false"
  | SFliteral(l) -> l
  | SBinop(e1, o, e2) ->
      string_of_sexpr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_sexpr e2
  | SUnop(o, e) -> string_of_uop o ^ string_of_sexpr e
  | SCall(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_sexpr el) ^ ")"
  | SCreateNote(el) ->
      "new Note(" ^ String.concat ", " (List.map string_of_sexpr el) ^ ")"
  | SAssign(v, e) -> string_of_sexpr v ^ " = " ^ string_of_sexpr e
  | SId(s) -> s
  | SNoteAccess(s, p) -> string_of_sexpr s ^ "." ^ string_of_prop p
  | SNoteAssign(s, p, e) -> string_of_sexpr s ^ "." ^ string_of_prop p ^ " = " ^ string_of_sexpr e
  (*TODO: Make it say something more helpful?*)
  | SArrayLiteral(_) -> "ARRAY LITERAL"
  | SArrayAccess(_, _) -> "ARRAY ACCESSING"
  | SArrayLength(_) -> "ARRAY LENGTH"
  | SCreateArray(_, _) -> "CREATING A NEW ARRAY"
  | SNoexpr -> ""
				  ) ^ ")"

let  rec  string_of_sstmt = function
    SExpr(expr) -> string_of_sexpr expr ^ ";\n"
  | SBlock(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_sstmt stmts) ^ "}\n"
  | SReturn(expr) -> "return " ^ string_of_sexpr expr ^ ";\n"
  | SBind(t, id) -> string_of_typ t ^ " " ^ id ^ ";\n"
  | SBindAssign(t, id, e) -> string_of_typ t ^ " " ^ id ^ " = " ^ string_of_sexpr e ^ ";\n"
  | SIf(e, s, SBlock([])) ->
      "if (" ^ string_of_sexpr e ^ ")\n" ^ string_of_sstmt s
  | SIf(e, s1, s2) ->  "if (" ^ string_of_sexpr e ^ ")\n" ^
      string_of_sstmt s1 ^ "else\n" ^ string_of_sstmt s2
  | SFor(s1, e2, e3, s) ->
    "for (" ^ string_of_sstmt s1 ^ string_of_sexpr e2 ^ " ; " ^
    string_of_sexpr e3  ^ ") " ^ string_of_sstmt s
  | SWhile(e, s) -> "while (" ^ string_of_sexpr e ^ ") " ^ string_of_sstmt s


let string_of_sfdecl fdecl =
  string_of_typ fdecl.styp ^ " " ^
  fdecl.sfname ^ "(" ^ String.concat ", " (List.map snd fdecl.sformals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_sstmt fdecl.sbody) ^
  "}\n"

let string_of_sprogram (stmts, funcs) =
  String.concat "" (List.map string_of_sstmt stmts) ^ "\n" ^
  String.concat "\n" (List.map string_of_sfdecl funcs)
